import os
import math
import collections
import json
import shutil

from org.transcrypt import utils

'''
A cascaded mini mapping is made as follows:
    - First generated a non-cascaded pretty map
    - After that have the minifier generate a shrink map and load that
    - After that cascade the two to obtain the mini map and save that
'''

# Tools to embed source map info in target code

lineNrLength = 6
maxNrOfSourceLinesPerModule = 1000000

# Tools to encode and decode numbers as base 64 variable length quantities

class Base64VlqConverter:
    def __init__ (self):
        self.encoding = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'  # Forward lookup table, from index to character, 64 symbols in total
        self.decoding = dict ((char, i) for i, char in enumerate (self.encoding))           # Enable reverse lookup, so from character to index
        
        self.prefabSize = 256                                                               # The 256 most used vlq's are prefabricated
        self.prefab = [self.encode ([i], True) for i in range (self.prefabSize)]
    
    def encode (self, numbers, init = False):
        segment = ''
        for number in numbers:
            if not init and 0 < number < self.prefabSize:
                field = self.prefab [number]
            else:
                signed = bin (abs (number)) [2 : ] + ('1' if number < 0 else '0')           # Convert to binary, chop off '0b' and append sign bit
                nChunks = math.ceil (len (signed) / 5.0)                                    # Determine minimum nr of needed 5 bit chunks (2 ^ 5 == 32)
                padded = (5 * '0' + signed) [-nChunks * 5 : ]                               # Pad by prepending zeroes to fit integer nr of chunks
                chunks = [
                    ('1' if iChunk else '0') + padded [                                     # Prefix first chunk with 0, continuation chunks with 1 (continuation bit)
                        iChunk * 5 : (iChunk + 1) * 5                                       # Pick out a not yet prefixed chunk from the padded bits 
                    ] 
                    for iChunk in range (nChunks - 1, -1, -1)                               # Reverse chunks, so the chunk starting with 0 will be last, first has sign bit
                ]                                                                           # So encountering a chunk with int value < 32 will denote end of number
                field = ''.join ([self.encoding [int (chunk, 2)] for chunk in chunks])      # Convert each chunk, incl. continuation bit to its encoding
            segment += field
        return segment

    def decode (self, segment):
        numbers = []
        accu = 0
        weight = 1

        for char in segment:
            ordinal = self.decoding [char]
            isContinuation = ordinal >= 32
            
            if isContinuation:
                ordinal -=32                                                                # Reset continuation bit
                
            if weight == 1:                                                                 # It was the tail of a number
                sign = -1 if ordinal % 2 else 1                                             # Remember sign
                ordinal //= 2                                                               # Remove sign bit, no matter what it was
                
            accu += weight * ordinal                                                        # Add new ordinal as currently least significant
            
            if isContinuation:                                                              # If it's a continuation
                if weight == 1:                                                             #   If it's the first continuation it will have the sign bit
                    weight = 16                                                             #       So next weight is 16
                else:                                                                       #   Else it won't have the sign bit:
                    weight *= 32                                                            #       So next weight * 32
            else:                                                                           # Else  ('no continuation' means 'end of number', since chunks are reversed)
                numbers.append (sign * accu)                                                #   Append accumulated number to results
                accu = 0                                                                    #   Reset accumulator for next number
                weight = 1                                                                  #   Reset weight, next number will again start with least significant part
                
        return numbers
        
base64VlqConverter = Base64VlqConverter ()

# Tools to create and combine sourcemaps

mapVersion = 3
iTargetLine, iTargetColumn, iSourceIndex, iSourceLine, iSourceColumn = range (5)    # Line indexes rather than line numbers are stored  

class SourceMapper: # There's only one sourcemapper needed to generate all maps of a module
    def __init__ (
        self,
        moduleName,
        targetDir,
        minify,
        dump
    ):
        self.moduleName = moduleName
        self.targetDir = targetDir
        self.minify = minify
        self.dump = dump
        
    def generateAndSavePrettyMap (self, sourceLineNrs):
        self.prettyMappings = [[targetLineIndex, 0, 0, sourceLineNr - 1, 0] for targetLineIndex, sourceLineNr in enumerate (sourceLineNrs)]
        self.prettyMappings.sort ()
        
        infix = '.pretty'  if self.minify else ''
        
        self.save (self.prettyMappings, infix)
        
        if self.dump:
            self.dumpMap (self.prettyMappings, infix, '.py')
            self.dumpDeltaMap (self.prettyMappings, infix)
            
    def cascadeAndSaveMiniMap (self):
        def getCascadedMapping (shrinkMapping):                                 # N.B. self.prettyMappings has to be sorted in advance
            prettyMapping = self.prettyMappings [min (shrinkMapping [iSourceLine], len (self.prettyMappings) - 1)] 
            
            result =            (
                shrinkMapping [ : iTargetColumn + 1]                            # Target location from shrink mapping
                +
                prettyMapping [iSourceIndex : ]                                 # Source location from self
            )
            if self.dump:
                self.cascadeMapdumpFile.write ('{} {} {}\n'.format (result, shrinkMapping, prettyMapping))
            return result
        
        if self.dump:
            self.cascadeMapdumpFile = utils.create (f'{self.targetDir}/{self.moduleName}.cascade_map_dump')
        
        self.miniMappings = [
            getCascadedMapping (shrinkMapping)
            for shrinkMapping in self.shrinkMappings
        ]
        self.miniMappings.sort ()
        
        self.save (self.miniMappings, '')
        
        if self.dump:
            self.cascadeMapdumpFile.close ()
        
    def loadShrinkMap (self):
        with open  (f'{self.targetDir}/{self.moduleName}.shrink.map') as mapFile:
            rawMap = json.loads (mapFile.read ())
                
        deltaMappings = [
            [base64VlqConverter.decode (segment) for segment in group.split (',')]
            for group in rawMap ['mappings'] .split (';')
        ]
        
        '''
        Fields in a delta segment as directly decoded from the output of the minifier:
          index (target line index implicit, is group index)
            0: target column index
            1: source file index        (optional)      (always zero)
            2: source line index        (optional)
            3: source column index      (optional)
            4: name index               (optional)
            
        Fields in a shrinkMapping:
          index
            0: target line index (is group index, a group represents a target line)
            1: target column index
            2: source file index        (always zero)   (i = 1)
            3: source line index                        (i = 2)
            4: source column index                      (i = 3)
            5: source name index        (left out)
        '''
        
        self.shrinkMappings = []
        for groupIndex, deltaGroup in enumerate (deltaMappings):
            for segmentIndex, deltaSegment in enumerate (deltaGroup):
                if deltaSegment:                                                    # Shrink map ends with empty group, i.e. 'holding empty segment'
                    if segmentIndex:
                        self.shrinkMappings.append ([groupIndex, deltaSegment [0] + self.shrinkMappings [-1][1]])
                    else:                                                           # Start of group
                        self.shrinkMappings.append ([groupIndex, deltaSegment [0]]) # Absolute target column
                        
                    for i in range (1, 4):                                          # So i in [1, 2, 3]
                        if groupIndex or segmentIndex:
                            self.shrinkMappings [-1] .append (deltaSegment [i] + self.shrinkMappings [-2][i + 1])
                        else:                                                       # Start of map
                            try:
                                self.shrinkMappings [-1] .append (deltaSegment [i]) # Absolut file index, source line and source column
                            except:                                                 # Shrink map starts with 'A' rather than 'AAAA'
                                self.shrinkMappings [-1] .append (0)                       
        self.shrinkMappings.sort () # Sort on target line and inside that on target column
                                
        if self.dump:
            self.dumpMap (self.shrinkMappings, '.shrink', '.py')
            self.dumpDeltaMap (deltaMappings, '.shrink')
            
    def save (self, mappings, infix):
        deltaMappings = []
        oldMapping = [-1, 0, 0, 0, 0]
        for mapping in mappings:
            newGroup = mapping [iTargetLine] != oldMapping [iTargetLine]
            
            if newGroup:
                deltaMappings.append ([])                                                              # Append new group
                
            deltaMappings [-1] .append ([])                                                            # Append new segment, one for each mapping
            
            if newGroup:
                deltaMappings [-1][-1] .append (mapping [iTargetColumn])                               # Only target column reset for every group
            else:
                deltaMappings [-1][-1] .append (mapping [iTargetColumn] - oldMapping [iTargetColumn])  # Others are delta's, so cumulative
                    
            for i in [iSourceIndex, iSourceLine, iSourceColumn]:
                deltaMappings [-1][-1] .append (mapping [i] - oldMapping [i])
                    
            oldMapping = mapping
                                    
        rawMap = collections.OrderedDict ([
            ('version', mapVersion),
            ('file', f'{self.moduleName}.js'), # Target
            ('sources', [f'{self.moduleName}{infix}.py']),
            # ('sourcesContent', [None]),
            ('mappings', ';'.join ([
                ','.join ([
                    base64VlqConverter.encode (segment)
                    for segment in group
                ])
                for group in deltaMappings
            ]))
        ])
                
        with utils.create (f'{self.targetDir}/{self.moduleName}{infix}.map') as mapFile:
            mapFile.write (json.dumps (rawMap, indent = '\t'))
            
        if self.dump:
            self.dumpMap (mappings, infix, '.py')
            self.dumpDeltaMap (deltaMappings, infix)
            
    def dumpMap (self, mappings, infix, sourceExtension):
        with utils.create (f'{self.targetDir}/{self.moduleName}{infix}.map_dump') as mapdumpFile:
            mapdumpFile.write (f'mapVersion: {mapVersion}\n\n')
            mapdumpFile.write (f'targetPath: {self.moduleName}.js\n\n')
            mapdumpFile.write (f'sourcePath: {self.moduleName}{infix}{sourceExtension}\n\n')
            mapdumpFile.write ('mappings:\n')
            for mapping in mappings:
                mapdumpFile.write ('\t{}\n'.format (mapping))
                
    def dumpDeltaMap (self, deltaMappings, infix):
        with utils.create (f'{self.targetDir}/{self.moduleName}{infix}.delta_map_dump') as deltaMapdumpFile:
            for group in deltaMappings:
                deltaMapdumpFile.write ('(New group) ')
                for segment in group:
                    deltaMapdumpFile.write ('Segment: {}\n'.format (segment))
                                
    def generateMultilevelMap (self):
        utils.log (False, 'Saving multi-level sourcemap in: {}\n') # !!!     , 'self.mapPath')
        self.loadShrinkMap ()
        self.cascadeAndSaveMiniMap ()
        


import asyncio

import pytest

from pret.store import create_store


@pytest.mark.asyncio
async def test_sync_two_states_update(tmp_path):
    path = tmp_path / "state.bin"
    state1 = create_store({"value": 0}, sync=path)
    state2 = create_store({"value": 0}, sync=path)
    assert state1.doc.sync_id == state2.doc.sync_id

    await asyncio.sleep(0.01)

    state1["value"] = 1
    await asyncio.sleep(0.1)
    await asyncio.sleep(0.1)
    assert state1["value"] == 1
    assert state2["value"] == 1

    state2["value"] = 2
    await asyncio.sleep(0.5)
    assert state1["value"] == 2

    state1.doc._persistence_watcher.cancel()
    state2.doc._persistence_watcher.cancel()
    await asyncio.sleep(0)
    state1.doc._persistence_finalizer.detach()
    state2.doc._persistence_finalizer.detach()
    state1.doc._persistence_finalizer = None
    state2.doc._persistence_finalizer = None


@pytest.mark.asyncio
async def test_hydrate_from_existing_file(tmp_path):
    path = tmp_path / "persist.bin"
    state1 = create_store({"count": 0}, sync=path)
    state1["count"] = 7
    await asyncio.sleep(0.05)
    state1.doc._persistence_watcher.cancel()
    await asyncio.sleep(0)
    state1.doc._persistence_finalizer.detach()
    state1.doc._persistence_finalizer = None

    state2 = create_store({}, sync=path)
    assert state2.doc.sync_id == state1.doc.sync_id
    await asyncio.sleep(0.05)
    assert state2["count"] == 7
    state2.doc._persistence_watcher.cancel()
    await asyncio.sleep(0)
    state2.doc._persistence_finalizer.detach()
    state2.doc._persistence_finalizer = None



import asyncio
import inspect
import mimetypes
from pathlib import Path
from typing import Dict, Union

from quart import (
    Quart,
    Response,
    abort,
    request,
    send_file,
    send_from_directory,
    websocket,
)

from pret.manager import get_manager


def make_app(assets: Dict[str, Union[str, Path]]) -> Quart:
    app = Quart(__name__)

    manager = get_manager()

    websocket_id = 0

    @app.websocket("/ws")
    async def ws() -> None:
        nonlocal websocket_id
        websocket_id += 1

        send_queue = manager.register_connection(websocket_id)

        async def _send():
            while True:
                await websocket.send_json(await send_queue.get())

        task = None
        try:
            task = asyncio.ensure_future(_send())

            while True:
                message = manager.handle_websocket_msg(
                    await websocket.receive_json(),
                    websocket_id,
                )
                await websocket.send_json(message)
        finally:
            if task is not None:
                task.cancel()
            manager.unregister_connection(websocket_id)

    @app.route("/method", methods=["GET", "POST"])
    async def method_route():
        body = await request.get_json()
        method, data = body["method"], body["data"]
        result = manager.handle_message(method, data)
        if inspect.isawaitable(result):
            result = await result
        if result is not None:
            return {"method": result[0], "data": result[1]}
        else:
            return {}

    @app.route("/", defaults={"path": "index.html"})
    @app.route("/<path:path>")
    async def page_route(path):
        mimetype = mimetypes.guess_type(path)[0]
        if path in assets:
            asset = assets[path]
            if isinstance(asset, Path):
                return await send_file(asset, mimetype=mimetype)
            elif isinstance(asset, str):
                return Response(asset, mimetype=mimetype)
            elif isinstance(asset, bytes):
                return Response(asset, mimetype="application/octet-stream")
        else:
            try:
                return await send_from_directory(assets["*"], path)
            except (KeyError, FileNotFoundError):
                abort(404)  # If there's no default file, return a 404 error

    return app



"""
This module provides marshaling to convert Python objects (and functions, and classes)
into a format that can be serialized and later reconstructed in a JavaScript
environment.
"""

import ast
import base64
import collections
import contextlib
import functools
import inspect
import os
import pathlib
import re
import sys
import tempfile
import textwrap
import time
import types
import uuid
import weakref
from asyncio import Future
from io import BytesIO, StringIO
from pathlib import Path
from types import FunctionType
from typing import Any, Callable, Dict, List, Optional, Union
from weakref import WeakKeyDictionary

import astunparse
from transcrypt.__main__ import main as transcrypt

try:
    import cbor2._encoder as cbor2_encoder
except ImportError:
    # Python 3.7 and earlier do not have cbor2._encoder
    import cbor2.encoder as cbor2_encoder

ModuleType = type(sys)
BuiltinFunctionType = type(time.time)
marshal_overrides = WeakKeyDictionary()
_current_marshaller: Optional["PretMarshaler"] = None


def marshal_as(
    base_version: Any = None,
    marshaled_version: Any = None,
    js: str = None,
    globals: Union[dict, Callable[[], Dict]] = {},
):
    @functools.wraps(marshal_as)
    def wrap(fn):
        assert fn is not None
        if marshaled_version is not None:
            assert not isinstance(marshaled_version, str)
            marshalable = ("py", marshaled_version)
        else:
            assert js is not None
            if callable(globals):
                scoped_vars = globals()
            else:
                scoped_vars = globals
            marshalable = ("js", js, scoped_vars)
        marshal_overrides[fn] = marshalable
        return fn

    if base_version is None:
        return wrap
    if marshaled_version is None and js is None:
        marshaled_version = base_version
        return wrap

    return wrap(base_version)


class GlobalRef:
    def __init__(self, module, name):
        self.module = module
        self.name = name
        self.__module__ = module.__name__

    def __reduce__(self):
        global _current_marshaller
        # str is interpreted by pickle as save global, which is exactly what we want
        _current_marshaller.accessed_global_refs.add(self)
        return self.__module__ + "." + self.name

    def __hash__(self):
        return hash(self.__module__ + "." + self.name)

    def __repr__(self):
        return f"GlobalRef({self.__module__}.{self.name})"

    def __str__(self):
        return f"{self.__module__}.{self.name}"


js = ModuleType("js", None)
marshal_as(js, js="return window.pret_modules.js")


def make_stub_js_module(
    global_name,
    py_package_name,
    js_package_name,
    package_version,
    stub_qualified_name,
):
    # Makes a dummy module with __name__ set to the module name
    # so that it can be pickled and unpickled
    full_global_name = "js." + global_name

    def make_stub_js_function(name):
        # Makes a dummy function with __module__ set to the module name
        # so that it can be pickled and unpickled
        ref = GlobalRef(module, name)
        setattr(module, name, ref)
        return ref

    module = ModuleType(
        full_global_name, f"Fake server side js module for {global_name}"
    )
    module.__file__ = f"<{full_global_name}>"
    module.__getattr__ = make_stub_js_function
    module._js_package_name = js_package_name
    module._package_name = py_package_name
    module._package_version = package_version
    module._stub_qualified_name = stub_qualified_name
    sys.modules[module.__name__] = module
    marshal_as(module, js=f"return pret_modules.js.{global_name};")
    setattr(js, global_name, module)


@contextlib.contextmanager
def redirect_argv(*args):
    sys._argv = sys.argv[:]
    sys.argv = list(args)
    yield
    sys.argv = sys._argv


@contextlib.contextmanager
def capture_stdout():
    new_target = StringIO()
    old_target = sys.stdout
    sys.stdout = new_target
    try:
        yield new_target
    finally:
        sys.stdout = old_target


marshal_as(
    Future,
    js="""
import {Exception} from './org.transcrypt.__runtime__.js';

var CancelledError = _class_ ('CancelledError', [Exception], {
    __module__: __name__,
    get __init__() {return __get__(this, function(self, message) {
        Error.call(self, message || 'Future was cancelled');
        self.name = 'CancelledError';
        self.message = message || 'Future was cancelled';
    });}
});

var Future = _class_ ('Future', [object], {
    __module__: __name__,
    // States
    PENDING: 0,
    FINISHED: 1,
    CANCELLED: 2,

    get __init__() {return __get__(this, function(self) {
        self._state = self.PENDING;
        self._result = undefined;
        self._exception = undefined;
        self._promise = new Promise(function(resolve, reject) {
            self._resolve = resolve;
            self._reject = reject;
        });
    });},

    // helpers for awaiting
    get then() {return __get__(this, function(self, onFulfilled, onRejected) {
        return self._promise.then(onFulfilled, onRejected);
    });},
    get catch() {return __get__(this, function(self, onRejected) {
        return self._promise.catch(onRejected);
    });},
    get finally() {return __get__(this, function(self, onFinally) {
        return self._promise.finally(onFinally);
    });},

    // properties
    get done() {return __get__(this, function(self) {
        return self._state !== self.PENDING;
    });},
    get cancelled() {return __get__(this, function(self) {
        return self._state === self.CANCELLED;
    });},
    get result() {return __get__(this, function(self) {
        if (!self.done()) throw new Error('Future not done yet');
        if (self.cancelled()) throw new Error('Future was cancelled');
        if (self._exception !== undefined) throw self._exception;
        return self._result;
    });},
    get exception() {return __get__(this, function(self) {
        return self.done() ? self._exception : undefined;
    });},

    // mutators
    get set_result() {return __get__(this, function(self, value) {
        if (self.done()) return false;
        self._state = self.FINISHED;
        self._result = value;
        self._resolve(value);
        return true;
    });},
    get set_exception() {return __get__(this, function(self, err) {
        if (self.done()) return false;
        self._state = self.FINISHED;
        self._exception = err instanceof Error ? err : new Error(String(err));
        self._reject(self._exception);
        return true;
    });},
    get cancel() {return __get__(this, function(self, msg) {
        if (self.done()) return false;
        self._state = self.CANCELLED;
        self._exception = CancelledError(msg || 'Future was cancelled');
        self._reject(self._exception);
        return true;
    });}
});

Future.CancelledError = CancelledError;

return Future;
""",
)


PRE_TRANSPILED_MODULES = {
    "builtins": "org.transcrypt.__runtime__",
    "cmath": "cmath",
    "copy": "copy",
    "dataclasses": "dataclasses",
    "datetime": "datetime",
    "functools": "functools",
    "itertools": "itertools",
    "json": "json",
    "logging": "logging",
    "math": "math",
    "random": "random",
    "re": "re",
    "time": "time",
    "typing": "typing",
    "unicodedata": "unicodedata",
    "warnings": "warnings",
    "weakref": "weakref",
}


def inspect_scopes(f):
    global_vars = set()
    stack = [f.__code__]
    while stack:
        code = stack.pop()
        stack.extend(k for k in code.co_consts if isinstance(k, types.CodeType))
        global_vars |= {n for n in code.co_names}

    globals_vals = {
        n: f.__globals__[n] for n in sorted(global_vars) if n in f.__globals__
    }
    closure_vals = (
        {}
        if f.__closure__ is None
        else {
            n: cell.cell_contents
            for n, cell in zip(f.__code__.co_freevars, f.__closure__)
        }
    )
    return {**globals_vals, **closure_vals}


class StrictCBOREncoder(cbor2_encoder.CBOREncoder):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        for base, marshaled in marshal_overrides.items():
            if marshaled[0] == "py" and marshaled[1] in self._encoders:
                # If we have a py marshalable, we can use it directly
                self._encoders[base] = self._encoders[marshaled[1]]

    def _find_encoder(
        self, obj_type: type
    ) -> "Callable[[cbor2_encoder.CBOREncoder, Any], None] | None":
        for type_or_tuple, enc in list(self._encoders.items()):
            if type(type_or_tuple) is tuple:
                try:
                    modname, typename = type_or_tuple
                except (TypeError, ValueError):
                    raise cbor2_encoder.CBOREncodeValueError(
                        f"invalid deferred encoder type {type_or_tuple!r} (must be a "
                        "2-tuple of module name and type name, e.g. "
                        "('collections', 'defaultdict'))"
                    )

                imported_type = getattr(sys.modules.get(modname), typename, None)
                if imported_type is not None:
                    del self._encoders[type_or_tuple]
                    self._encoders[imported_type] = enc
                    type_ = imported_type
                else:  # pragma: nocover
                    continue
            else:
                type_ = type_or_tuple

            if obj_type is type_:
                self._encoders[obj_type] = enc
                return enc
        return None

    def encode(self, obj: Any) -> None:
        obj_type = obj.__class__
        if obj_type in self._encoders:
            encoder = self._encoders[obj_type] or self._default
        else:
            encoder = self._find_encoder(obj_type) or self._default
        if not encoder:
            raise cbor2_encoder.CBOREncodeTypeError(
                f"cannot serialize type {obj_type.__name__}"
            )

        encoder(self, obj)


class PretMarshaler:
    def __init__(self, allow_error: bool = False):
        self.source_codes = {}
        self.allow_error = allow_error
        self.accessed_global_refs = set()
        self.encoder = cbor2_encoder.shareable_encoder(self._encoder)
        self.id = uuid.uuid4().hex
        self.chunk_idx = 0
        self._file = BytesIO()
        self._cbor_encoder = StrictCBOREncoder(
            self._file, default=self.encoder, value_sharing=True
        )

    def transpile(self):
        py_module_string = (
            "__pragma__ ('kwargs')\n"
            + "\n".join(v[1] for k, v in self.source_codes.items() if v[0] == "py")
            + "\n__pragma__ ('nokwargs')\n"
        )
        tmpdir = Path(tempfile.mkdtemp())
        os.makedirs(tmpdir / "__target__", exist_ok=True)
        with open(tmpdir / "__main__.py", "w") as f:
            f.write(py_module_string)
        with redirect_argv(
            *("-b", "-n", "-m", "-p", "window"),
            str(tmpdir / "__main__.py"),
        ):
            with capture_stdout() as output:
                transcrypt()
            if "Saving" not in output.getvalue() or "Error" in output.getvalue():
                raise Exception(
                    "Could not export your Python app to Javascript with Transcrypt. "
                    "Here is the log: \n{}".format(output.getvalue())
                )
        res = (tmpdir / "__target__/__main__.js").read_text()
        res += "\n" + "\n".join(
            v[1] for k, v in self.source_codes.items() if v[0] == "js"
        )
        for k, v in self.source_codes.items():
            assert v[1].strip() != "", f"Source code for {k} is empty: {v}"
        (tmpdir / "__target__/__main__.js").write_text(res)
        return tmpdir

    def process_js_module(self, js_module):
        has_default = False
        if "export default " in js_module:
            js_module = js_module.replace("\nexport default ", "\nvar _default_ = ")
            has_default = True
        exported = [
            match.group(1) or match.group(2).strip()
            for match in re.finditer(
                r"\nexport\s+(?:(?:var|function|let|const)?\s+([\w_]+)|{([^}]*)})",
                js_module,
            )
        ]
        imports = []
        js_module = js_module.replace("\nexport ", "")

        if has_default:
            exported.append("_default_")
        js_module += "\nreturn {" + ", ".join(exported) + "};"

        def replace_js_import(match):
            imported = match.group(1).strip()
            mod_key = match.group(2).strip()
            if mod_key.endswith(".js"):
                mod_key = mod_key[:-3]
            if mod_key.startswith("./"):
                mod_key = mod_key[2:]
            mod_key = mod_key.replace("/", ".")

            imports.append(mod_key)

            non_bracket_part, named_part = (
                (None, imported)
                if imported.startswith("{")
                else (
                    imported.split(",", 1)[0].strip(),
                    imported.split(",", 1)[1].strip() if "," in imported else None,
                )
            )

            lines = []
            if non_bracket_part:
                if non_bracket_part.startswith("* as "):
                    non_bracket_part = non_bracket_part[4:].strip()
                    lines.append(f"var {non_bracket_part} = pret_modules['{mod_key}'];")
                else:
                    lines.append(
                        f"var {non_bracket_part} = pret_modules['{mod_key}']._default_;"
                    )

            if named_part:
                inner = named_part[1:-1].strip()
                specs = []
                for s in (s.strip() for s in inner.split(",") if s.strip()):
                    specs.append(
                        f"{s.split(' as ')[0].strip()}: {s.split(' as ')[1].strip()}"
                        if " as " in s
                        else s
                    )
                lines.append(
                    f"var {{ {', '.join(specs)} }} = pret_modules['{mod_key}'];"
                )

            return "\n".join(lines)

        js_module = re.sub(
            "^import\\s+([^;]+)\\s+from\\s+['\"]([^'\"]+)['\"];",
            replace_js_import,
            js_module,
            flags=re.MULTILINE,
        )
        return js_module, imports

    def visit(self, obj):
        global _current_marshaller
        _current_marshaller = self
        self._cbor_encoder.encode(obj)
        _current_marshaller = None

    def dump(self, obj):
        chunk_idx = self.chunk_idx
        self.visit(obj)
        blob = self._file.getvalue()
        self.chunk_idx += 1
        header = (
            "if(window.pret_modules===undefined){window.pret_modules={};}\n"
            "var pret_modules=window.pret_modules;\n"
            "pret_modules.js=window;\n"
        )
        tmpdir = self.transpile()
        mods, deps = {}, {}
        for p in pathlib.Path(tmpdir / "__target__").glob("**/*.js"):
            code, imps = self.process_js_module(p.read_text())
            module_name = p.stem
            if module_name == "__main__":
                mods[module_name] = code
            else:
                mods[module_name] = (
                    f"pret_modules['{module_name}']=(function(){{{code}}})();"
                )
            deps[module_name] = {pathlib.Path(i).name for i in imps}

        indeg = {m: sum(d in mods for d in deps[m]) for m in mods}
        queue = collections.deque([m for m, n in indeg.items() if n == 0])
        order = []
        while queue:
            m = queue.popleft()
            order.append(m)
            for n in mods:
                if m in deps[n]:
                    indeg[n] -= 1
                    if indeg[n] == 0:
                        queue.append(n)
        order += [m for m in mods if m not in order]  # fallback for cycles
        order.remove("org.transcrypt.__runtime__")  # already bundled in js

        js = (
            header
            + "\n\n".join(mods[m] for m in order)
            + "\n//# sourceURL=dynamic_factory.js"
        )
        return [base64.encodebytes(blob).decode(), js], chunk_idx

    def _encoder(self, encoder, value):
        source_code_id = f"pret_factory_{len(self.source_codes)}"
        try:
            if value in marshal_overrides:
                marshalable = marshal_overrides[value]
                marshal_mode = marshalable[0]
                target = marshalable[1]
                if marshal_mode == "py":
                    # Python marshalable, we can just use the value
                    self.encoder(encoder, target)
                    return
                elif marshal_mode == "js":
                    # JS marshalable, we need to encode it as a tag
                    globals = marshalable[2] if len(marshalable) > 2 else {}
                    args = ", ".join(globals.keys())
                    factory_code = (
                        f"function {source_code_id}({args})"
                        " {\n" + f"{target.strip()};\n" + "}\n"
                        f"export {{{source_code_id}}};\n"
                    )
                    self.source_codes[source_code_id] = ("js", factory_code)
                    encoder.encode(
                        cbor2_encoder.CBORTag(
                            4000, [source_code_id, list(globals.values())]
                        )
                    )
                    return
            elif getattr(value, "__module__", None) in PRE_TRANSPILED_MODULES:
                # If the function is from a pre-transpiled module, we can just use its
                # name.
                factory_code = (
                    f"def {source_code_id}():\n"
                    f"  from {PRE_TRANSPILED_MODULES[value.__module__]} import {value.__name__}\n"  # noqa: E501
                    f"  return {value.__name__}\n"
                )
                self.source_codes[source_code_id] = ("py", factory_code)
                encoder.encode(cbor2_encoder.CBORTag(4000, [source_code_id, []]))
                return
            elif (
                isinstance(value, ModuleType)
                and value.__name__ in PRE_TRANSPILED_MODULES
            ):
                factory_code = (
                    f"def {source_code_id}():\n"
                    f"  import {PRE_TRANSPILED_MODULES[value.__name__]} as module\n"
                    f"  return module\n"
                )
                self.source_codes[source_code_id] = ("py", factory_code)
                encoder.encode(cbor2_encoder.CBORTag(4000, [source_code_id, []]))
                return
            elif isinstance(value, (FunctionType, BuiltinFunctionType)):
                # code = pygetsource.getfactory(value.__code__, function_name="_fn_")
                code = inspect.getsource(value.__code__)
                code = textwrap.dedent(code)
                scoped_vars = inspect_scopes(value)
                if "__class__" in scoped_vars:
                    # do assert here ?
                    scoped_vars.pop("__class__")
                # Factory will be used to create the fn or cls
                tree = ast.parse(code)
                tree.body[0].decorator_list = []
                factory = ast.FunctionDef(
                    name=source_code_id,
                    args=ast.arguments(
                        args=[ast.arg(arg=n, annotation=None) for n in scoped_vars],
                        kwarg=None,
                        vararg=None,
                        defaults=[],
                    ),
                    body=[
                        tree,
                        ast.Return(value=ast.Name(id=value.__name__, ctx=ast.Load())),
                    ],
                    decorator_list=[],
                )
                ast.fix_missing_locations(factory)
                factory_code = astunparse.unparse(factory)
                self.source_codes[source_code_id] = ("py", factory_code)
                encoder.encode(
                    cbor2_encoder.CBORTag(
                        4000, [source_code_id, list(scoped_vars.values())]
                    )
                )
                return
            elif isinstance(value, type):
                # 1. Get own members of the class (members that are not inherited)
                base = value.__base__
                members_names = sorted(
                    set(dir(value))
                    - {
                        "__dict__",
                        "__weakref__",
                        "__slots__",
                        "__subclasshook__",
                        "__init_subclass__",
                        "__abstractmethods__",
                    }
                )
                base_members_names = sorted(
                    set(dir(base))
                    - {
                        "__dict__",
                        "__weakref__",
                        "__slots__",
                        "__subclasshook__",
                        "__init_subclass__",
                        "__abstractmethods__",
                    }
                )
                members = [(name, getattr(value, name)) for name in members_names]
                parent_members = (
                    [(n, getattr(base, n)) for n in base_members_names] if base else []
                )
                own_members = [
                    member
                    for member in members
                    if member not in parent_members and member[0]
                ]
                methods = [
                    member
                    for member in own_members
                    if isinstance(member[1], FunctionType)
                ]
                encoder.encode(
                    cbor2_encoder.CBORTag(
                        4001,
                        [
                            value.__name__,
                            value.__bases__,
                            {n: m for n, m in own_members if n not in methods},
                            {n: m for n, m in methods if n not in own_members},
                        ],
                    )
                )
                return
            elif inspect.isclass(type(value)):
                state: Dict = None
                if isinstance(value, ModuleType):
                    # Do we really need to raise here ? Why not just encode the module
                    # as a dict ? Was this error raised to check that I got no module to
                    # marshal in some tests ?
                    # raise ValueError(value)
                    state = {
                        k: v for k, v in value.__dict__.items() if not is_builtins(k, v)
                    }
                if (
                    hasattr(value, "__reduce__")
                    and type(value).__reduce__ is not object.__reduce__
                ):
                    # If the object has a __reduce__ method,
                    # we can use it to get the state
                    reduced = value.__reduce__()
                    if isinstance(reduced, str):
                        factory_code = (
                            f"function {source_code_id}() {{\n"
                            f"  return pret_modules.{reduced};\n"
                            "}\n"
                            f"export {{{source_code_id}}};\n"
                        )
                        self.source_codes[source_code_id] = ("js", factory_code)
                        encoder.encode(
                            cbor2_encoder.CBORTag(4000, [source_code_id, []])
                        )
                        return
                    else:
                        encoder.encode(cbor2_encoder.CBORTag(4002, reduced))
                        return
                else:
                    try:
                        if state is None:
                            state = value.__dict__
                    except AttributeError:
                        state = {}
                    encoder.encode(
                        cbor2_encoder.CBORTag(
                            4003, [type(value), state, type(value).__name__]
                        )
                    )
                    return
            raise ValueError()
        except (RecursionError, TypeError):
            import traceback

            if self.allow_error:
                return  # self.encoder(encoder, None)

            else:
                traceback.print_exc()
                raise ValueError(
                    f"Could not marshal {value} of type {type(value)}. "
                    "Please ensure it is a marshalable type."
                )
        except BaseException:
            if self.allow_error:
                return  # self.encoder(encoder, None)
            else:
                raise


def is_builtins(key, value):
    # TODO expand ? or refacto ?
    # This is a small utility to check whether something is definitely not
    # marshalable, so we can skip it.
    if key.startswith("__") and key.endswith("__"):
        return True
    if isinstance(value, (BuiltinFunctionType, types.BuiltinFunctionType)):
        return True
    if isinstance(value, (type(Any), type(List))):
        return True
    if isinstance(value, ModuleType) and value.__name__ in sys.builtin_module_names:
        return True
    return False


shared_marshaler: Any = None


def get_shared_marshaler():
    global shared_marshaler
    if shared_marshaler is None or shared_marshaler() is None:
        marshaler = PretMarshaler()
        shared_marshaler = weakref.ref(marshaler)
    return shared_marshaler()


def clear_shared_marshaler():
    global shared_marshaler
    shared_marshaler = None



import asyncio
import contextlib
import fnmatch
import hashlib
import json
import os
import shutil
import subprocess
import sys
import tempfile
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from pret.marshal import get_shared_marshaler
from pret.serve import make_app


class RawRepr(str):
    def __repr__(self):
        return self


class BundleMode(str, Enum):
    FEDERATED = "federated"
    MONOLITHIC = "monolithic"


@contextlib.contextmanager
def build(
    renderables,
    static_dir: Union[str, Path] = None,
    build_dir: Union[str, Path] = None,
    mode: Union[bool, str, BundleMode] = True,
    dev: bool = False,
) -> Tuple[Dict[str, Union[str, Path]], List[Tuple[str, str]], str]:
    """
    Build the Pret app, pooling all the assets and entry points from
    the packages that were accessed for rendering.

    Parameters
    ----------
    renderables: List[pret.render.Renderable]
        The list of renderables to be bundled
    static_dir: Union[str, Path]
        The directory where the static files will be stored
    build_dir: Union[str, Path]
        The directory where the build files will be stored
    mode: Union[bool, str, BundleMode]
        The mode to use for bundling :

        - "federated": The app will be bundled where the assets are
          just copied from each package. This means that there may
          be repeated assets in the final bundle.
        - "monolithic": The app will be bundled where a consolidated
          bundle is created with all the assets, using webpack.

        By default, the mode is "federated".
    dev: bool
        Only used when mode is "monolithic". If True, the bundle will
        be created in development mode. Otherwise, it will be created
        in production mode.

    Returns
    -------
    Tuple[Dict[str, Union[str, Path]], List[Tuple[str, str]], str]
        A tuple containing the assets, entries and pickle filename
    """
    npm = None
    if mode is True:
        mode = BundleMode.FEDERATED
        # npm = shutil.which("npm")
        # if not npm:
        #     if mode == BundleMode.MONOLITHIC:
        #         raise Exception("npm not found. Please install node to proceed.")
        #     else:
        #         mode = BundleMode.FEDERATED
        # elif mode != BundleMode.MONOLITHIC:
        #     mode = BundleMode.MONOLITHIC

    if static_dir is None:
        static_dir = Path(tempfile.mkdtemp())
        delete_static = True
    else:
        static_dir = Path(static_dir)
        delete_static = False

    if build_dir is None:
        build_dir = Path(tempfile.mkdtemp())
        delete_build = True
    else:
        build_dir = Path(build_dir)
        delete_build = False

    marshaler = get_shared_marshaler()
    marshaler_file_str = ""
    for renderable in renderables:
        marshaler_file_str = json.dumps(renderable.bundle()[0])

    # Extract js globals and them to a temp file to be bundled with webpack
    js_globals, packages = extract_js_dependencies(marshaler.accessed_global_refs)

    js_globals_file = build_dir / "globals.ts"
    content_hash = hashlib.md5(marshaler_file_str.encode("utf-8")).hexdigest()[:20]
    bundle_filename = f"bundle.{content_hash}.pkl"

    if mode == BundleMode.MONOLITHIC:
        with js_globals_file.open("w") as f:
            f.write(js_globals)

        with (static_dir / bundle_filename).open("w") as f:
            f.write(marshaler_file_str)

        webpack_config = Path(__file__).parent / "webpack.standalone.js"
        # fmt: off
        # run npm webpack ... in cwd:
        subprocess.check_call(
            [
                npm,
                "exec",
                "webpack",
                "--config", webpack_config,
                "--env", "pretGlobalsFile=" + str(js_globals_file),
                "--mode", "development" if dev else "production",
            ],
            cwd=os.getcwd(),
        )
        # fmt: on

        assets = {
            "*": static_dir,
        }

    entries: List[Tuple[str, Optional[str]]] = []
    if mode == BundleMode.FEDERATED:
        extension_static_mapping, entries = extract_prebuilt_extension_assets(packages)
        base_static_mapping, index_html = extract_prebuilt_base_assets()
        index_html_str = index_html.read_text()
        index_html_str = (
            index_html_str.replace(
                "/* PRET_HEAD_TAGS */",
                "".join(
                    '<script src="{}"></script>'.format(remote_entry_file)
                    for remote_entry_file, _ in entries
                ),
            )
            .replace(
                "'__PRET_REMOTE_IMPORTS__'",
                str([remote_name for _, remote_name in entries]),
            )
            .replace("__PRET_PICKLE_FILE__", bundle_filename)
        )

        assets = {
            **base_static_mapping,
            **extension_static_mapping,
            bundle_filename: marshaler_file_str,
            # override the index.html file in base_static_mapping
            "index.html": index_html_str,
        }

        # static_dir.mkdir(parents=True, exist_ok=True)
        # Include entry points in the index html file

    yield assets, entries, bundle_filename

    if delete_static:
        shutil.rmtree(static_dir)
    if delete_build:
        shutil.rmtree(build_dir)


def extract_js_dependencies(
    refs,
    exclude=(
        "js.React.*",
        "js.ReactDOM.*",
    ),
):
    """
    Create a js file that will import all the globals that were accessed during
    pickling and assign them to the global scope.

    Parameters
    ----------
    refs: List[pret.marshal.GlobalRef]
        List of Ref objects that were accessed during pickling
    exclude: List[str]
        List of module patterns to exclude from the js globals file

    Returns
    -------
    """

    exported = {}
    imports = defaultdict(list)
    packages = []

    for ref_idx, ref in enumerate(refs):
        if not ref.__module__.startswith("js.") or any(
            fnmatch.fnmatch(str(ref), x) for x in exclude
        ):
            continue

        js_module_path_parts = ref.name.split(".")
        packages.append(ref.module)

        imports[
            ".".join((ref.module._js_package_name, *js_module_path_parts[:-1]))
        ].append((js_module_path_parts[-1], f"i_{ref_idx}"))

        current = exported
        for part in (ref.__module__[3:], *js_module_path_parts[:-1]):
            if part not in current:
                current[part] = {}

            current = current[part]
        current[ref.name.split(".")[-1]] = RawRepr(f"i_{ref_idx}")

    js_file_string = ""

    for package, aliases in imports.items():
        aliases_str = ", ".join([obj + " as " + alias for obj, alias in aliases])
        js_file_string += f"import {{ {aliases_str} }} from '{package}';\n"

    js_file_string += "\n\n"

    for globalName, value in exported.items():
        js_file_string += f"(window as any).{globalName} = {repr(value)};\n"

    return js_file_string, packages


def extract_prebuilt_extension_assets(
    packages: List[str],
) -> Tuple[Dict[str, Path], List[Tuple[str, str]]]:
    """
    Extracts entry javascript files from the static directory of each package
    as well as a mapping entry -> file to know where to look for whenever the app
    asks for a chunk or an asset.

    Parameters
    ----------
    packages

    Returns
    -------
    Tuple[Dict[str, Path], List[Tuple[str, str]]]
    """
    mapping = {}
    entries = []
    for package in set(packages):
        try:
            # in case it's an editable install
            stub_root = Path(sys.modules[package._stub_qualified_name].__file__).parent
            static_dir = stub_root / "js-extension" / "static"
            entry = next(static_dir.glob("remoteEntry.*.js"))
        except StopIteration:
            # otherwise, it's a regular install
            js_package = package._js_package_name
            static_dir = (
                Path(sys.prefix) / f"share/jupyter/labextensions/{js_package}/static"
            )
            entry = next(static_dir.glob("remoteEntry.*.js"))
        remote_name = json.loads((static_dir.parent / "package.json").read_text())[
            "name"
        ]
        mapping[entry.name] = entry
        entries.append((entry.name, remote_name))

        for static_file in static_dir.glob("*"):
            if static_file.name not in mapping:
                mapping[static_file.name] = static_file

    return mapping, entries


def extract_prebuilt_base_assets() -> Tuple[Dict[str, Path], Path]:
    """
    Extracts the base index.html file as well as a mapping entry -> file to know where
    to look for whenever the app asks for a chunk or an asset.

    Parameters
    ----------

    Returns
    -------
    Tuple[Dict[str, Path], List[str]]
    """
    mapping = {}
    static_dir = Path(__file__).parent / "js-base"
    entry = next(static_dir.glob("index.html"))

    for static_file in static_dir.glob("*.js"):
        if static_file.name not in mapping:
            mapping[static_file.name] = static_file

    return mapping, entry


def run(
    renderable,
    static_dir: Optional[Union[str, Path]] = None,
    build_dir: Optional[Union[str, Path]] = None,
    bundle: Union[bool, str, BundleMode] = True,
    dev: bool = True,
    serve: bool = True,
    port: int = 5000,
):
    """
    Serve the app, after building the app if necessary.

    Parameters
    ----------
    renderable: pret.render.Renderable
        The renderable object to be served
    static_dir: Optional[Union[str, Path]]
        The directory where the static files will be stored
    build_dir: Optional[Union[str, Path]]
        The directory where the build files will be stored
    bundle: Union[bool, str, BundleMode]
        The mode to use for bundling :

        - "federated": The app will be bundled where the assets are
          just copied from each package. This means that there may
          be repeated assets in the final bundle.
        - "monolithic": The app will be bundled where a consolidated
          bundle is created with all the assets, using webpack.

        By default, the mode is "federated".
    dev: bool
        Only used when mode is "monolithic". If True, the bundle will
        be created in development mode. Otherwise, it will be created
        in production mode.
    serve: bool
        Whether to serve the app after building it.
    port: int
        The port to use for serving the app.
    """

    with (
        build(
            [renderable],
            static_dir=static_dir,
            build_dir=build_dir,
            mode=bundle,
        )
        if bundle
        else contextlib.nullcontext({"*": Path(static_dir)})
    ) as (assets, entries, bundle_filename):
        app = make_app(assets)
        if serve:
            app.run(debug=dev, port=port, loop=asyncio.get_event_loop())
        return app



class EnvAutoCleaner(object):
    def __init__(self, ip):
        self.shell = ip
        self.cell_mapping = {}
        self.dunder_indices = {"_": None, "__": None, "___": None}

    def pre_run_cell(self, info):
        shell = self.shell
        old_idx = self.cell_mapping.get(info.cell_id)
        if old_idx is not None:
            try:
                del shell.user_ns[f"_{old_idx}"]
            except KeyError:
                pass
            try:
                del shell.user_ns_hidden[f"_{old_idx}"]
            except KeyError:
                pass
            try:
                del shell.history_manager.output_hist[old_idx]
            except KeyError:
                pass
            if old_idx == self.dunder_indices["_"]:
                shell.user_ns["_"] = None
                shell.user_ns_hidden["_"] = None
                shell.displayhook._ = None
                if shell.displayhook.exec_result is not None:
                    shell.displayhook.exec_result.result = None
                if shell.last_execution_result is not None:
                    shell.last_execution_result.result = None
            elif old_idx == self.dunder_indices["__"]:
                shell.user_ns["__"] = None
                shell.user_ns_hidden["__"] = None
                shell.displayhook.__ = None
            elif old_idx == self.dunder_indices["___"]:
                shell.user_ns["___"] = None
                shell.user_ns_hidden["___"] = None
                shell.displayhook.___ = None

        self.cell_mapping[info.cell_id] = self.shell.execution_count

        # TODO: handle case where I comment the code of a cell
        #       and execute (ie erase output)

    def post_run_cell(self, result):
        if result.result is not None:
            # shift dunder indices
            self.dunder_indices = {
                "_": self.shell.execution_count - 1,
                "__": self.dunder_indices["_"],
                "___": self.dunder_indices["__"],
            }


try:
    from IPython import get_ipython

    cleaner = EnvAutoCleaner(get_ipython())
    get_ipython().events.register("pre_run_cell", cleaner.pre_run_cell)
    get_ipython().events.register("post_run_cell", cleaner.post_run_cell)
except (ImportError, AttributeError):
    pass



import typer

from .update_jupyter_config import update_jupyter_config
from .stub_command import stub
from .prepack_command import prepack

app = typer.Typer()
app.command(name="stub")(stub)
app.command(name="prepack")(prepack)
app.command(name="update-jupyter-config")(update_jupyter_config)


@app.callback()
def callback():
    pass


if __name__ == "__main__":
    app()



# ruff: noqa: F821
import getpass
import os
from pathlib import Path
from tempfile import mkdtemp

import jupyterlab

# Test if we are running in a docker
if getpass.getuser() == "jovyan":
    c.ServerApp.ip = "0.0.0.0"  # noqa S104


c.ServerApp.port = 8889
c.ServerApp.port_retries = 0
c.ServerApp.open_browser = False
c.LabServerApp.extra_labextensions_path = str(
    Path(jupyterlab.__file__).parent / "galata"
)  # noqa: E501

c.LabApp.workspaces_dir = mkdtemp(prefix="galata-workspaces-")

c.ServerApp.root_dir = os.environ.get(
    "JUPYTERLAB_GALATA_ROOT_DIR", mkdtemp(prefix="galata-test-")
)
c.IdentityProvider.token = ""
c.ServerApp.token = ""  # For python 3.7, otherwise IdentityProvider.token is used
c.ServerApp.password = ""
c.ServerApp.disable_check_xsrf = True
c.LabApp.expose_app_in_browser = True



from typing import (
    Any,
    Callable,
    List,
    NewType,
    Optional,
    Tuple,
    TypeVar,
)

from typing_extensions import Protocol

from pret.marshal import js, marshal_as

# from pret.store import (
#     DictPretProxy,
#     ListPretProxy,
#     TrackedDictPretProxy,
#     TrackedListPretProxy,
# )

StateValueType = TypeVar("StateValueType")


@marshal_as(
    js="""
return window.React.useState
"""
)
def use_state(
    initial_value: "StateValueType",
) -> "Tuple[StateValueType, Callable[[StateValueType], None]]":
    """
    Returns a stateful value, and a function to update it.

    Examples
    --------

    ```python
    from pret.ui.react import div, button, p
    from pret import component, use_state


    @component
    def CounterApp():
        count, set_count = use_state(0)

        def increment():
            set_count(count + 1)

        return div(p(count), button({"onClick": increment}, "Increment"))
    ```

    Parameters
    ----------
    initial_value: StateValueType
        The initial value of the state

    Returns
    -------
    Tuple[StateValueType, Callable[[StateValueType], None]]

        - The current value of the state
        - A function to update the state
    """


T = TypeVar("T")


@marshal_as(
    js="""
return window.React.useMemo
"""
)
def use_memo(
    fn: "Callable[[], T]",
    dependencies: "List",
) -> "T":
    """
    Returns a memoized value, computed from the provided function.
    The function will only be re-executed if any of the dependencies change.

    !!! note

        Ensure that dependencies are simple values like int, str, bool
        to avoid unnecessary re-executions, as these values are converted to
        javascript objects, and converting complex objects may not ensure
        referential equality.

    Parameters
    ----------
    fn: Callable[[], T]
        The function to run to compute the memoized value
    dependencies: List
        The dependencies that will trigger a re-execution of the function

    Returns
    -------
    FunctionReturnType
        The value
    """


RefValueType = TypeVar("RefValueType")


class RefType(Protocol[RefValueType]):
    current: RefValueType


@marshal_as(
    js="""
return window.React.useRef
"""
)
def use_ref(initial_value: "RefValueType") -> "RefType[RefValueType]":
    """
    Returns a mutable ref object whose `.current` property is initialized to the
    passed argument.

    The returned object will persist for the full lifetime of the component.

    Parameters
    ----------
    initial_value: Any
        The initial value of the ref

    Returns
    -------
    RefType[RefValueType]
        The ref object
    """
    return js.React.useRef(initial_value)


C = NewType("C", Callable[..., Any])


@marshal_as(
    js="""
return window.React.useCallback
"""
)
def use_callback(
    callback: "C",
    dependencies: "Optional[List]" = None,
) -> "C":
    """
    Returns a memoized callback function. The callback will be stable across
    re-renders, as long as the dependencies don't change, meaning the last
    callback function passed to this function will be used between two re-renders.

    !!! note

        Ensure that dependencies are simple values like int, str, bool
        to avoid unnecessary re-executions, as these values are converted to
        JavaScript objects, and converting complex objects may not ensure
        referential equality.

    Parameters
    ----------
    callback: C
        The callback function
    dependencies: Optional[List]
        The dependencies that will trigger a re-execution of the callback.

    Returns
    -------

    """
    return js.React.useCallback(callback, dependencies)


def use_effect(effect: "Callable" = None, dependencies: "Optional[List]" = None):
    """
    The `useEffect` hook allows you to perform side effects in function components.
    Side effects can include data fetching, subscriptions, manually changing the DOM,
    and more.

    The effect runs after every render by default. If `dependencies` are provided,
    the effect runs whenever those values change. Therefore, if `dependencies` is an
    empty array, the effect runs only once after the initial render.

    !!! note

        Ensure that dependencies are simple values like int, str, bool
        to avoid unnecessary re-executions, as these values are converted to
        javascript objects, and converting complex objects may not ensure
        referential equality.

    Parameters
    ----------
    effect: Callable
        A function containing the side effect logic.
        It can optionally return a cleanup function.
    dependencies: Optional[List]
        An optional array of dependencies that determines when the effect runs.
    """
    if effect is None:

        def decorator(func):
            return window.React.useEffect(func, dependencies)  # noqa: F821

        return decorator
    # TODO: add window to scoped variables as `js`
    return window.React.useEffect(effect, dependencies)  # noqa: F821


def use_body_style(styles):
    def apply_styles():
        # Remember the original styles
        original_styles = {}
        for key, value in styles.items():
            original_styles[key] = getattr(js.document.documentElement.style, key, "")
            setattr(window.document.documentElement.style, key, value)  # noqa: F821

        # Cleanup function to revert back to the original styles
        def cleanup():
            for k, v in original_styles.items():
                setattr(window.document.documentElement.style, k, v)  # noqa: F821

        return cleanup

    use_effect(apply_styles, [styles])


# @overload
# def use_store_snapshot(
#     proxy_object: "Union[DictPretProxy, TrackedDictPretProxy]",
# ) -> "TrackedDictPretProxy": ...
#
#
# @overload
# def use_store_snapshot(
#     proxy_object: "Union[ListPretProxy, TrackedListPretProxy]",
# ) -> "TrackedListPretProxy": ...


@marshal_as(js="return window.storeLib.useSnapshot")
def use_store_snapshot(proxy_object):
    """
    This hook is used to track the access made on a store.
    You cannot use the returned object to change the store, you
    must mutate the original create_store(...) object directly.

    Parameters
    ----------
    proxy_object: ProxyType
        A store object, like the one returned by `create_store({...})`

    Returns
    -------
    TrackedProxyType
        A tracked store object
    """


@marshal_as(
    js="""
return function use_event_callback(callback) {
    const callbackRef = window.React.useRef(callback);
    callbackRef.current = callback;

    return window.React.useCallback(
        (function () {return callbackRef.current(...arguments)}),
        [],
    );
}
"""
)
def use_event_callback(callback: "C"):
    """
    This hook is used to store a callback function that will be called when an event
    is triggered. The callback function can be changed without triggering a re-render
    of the component. The function returns a wrapped callback function that will in
    turn call the stored callback function.

    !!! warning
        Do not use this hook if the rendering of the component depends on the callback
        function.

    Parameters
    ----------
    callback: C
        The callback function

    Returns
    -------
    C
        The wrapped callback function
    """



import os

from pret import component, create_store, use_state, use_store_snapshot
from pret.ui.react import div, input, label, p

state = create_store(
    {
        "faire à manger": True,
        "faire la vaisselle": False,
    },
    sync=os.environ["SYNC_FILE"],
)


@component
def TodoApp():
    todos = use_store_snapshot(state)
    typed, set_typed = use_state("")
    num_remaining = sum(not ok for ok in todos.values())
    plural = "s" if num_remaining > 1 else ""

    def on_key_down(event):
        if event.key == "Enter":
            state[typed] = False
            set_typed("")

    return div(
        *(
            div(
                input(
                    checked=ok,
                    type="checkbox",
                    id=todo.replace(" ", "-"),
                    on_change=lambda e, t=todo: state.update({t: e.target.checked}),
                ),
                label(todo, **{"for": todo.replace(" ", "-")}),
            )
            for todo, ok in todos.items()
        ),
        input(
            value=typed,
            on_change=lambda event: set_typed(event.target.value),
            on_key_down=on_key_down,
            placeholder="Add a todo",
        ),
        p(f"Number of unfinished todo{plural}: {num_remaining}", level="body-md"),
    )



import os
import sys
from pathlib import Path

import typer

from pret.marshal import PretMarshaler

sys.path.insert(0, os.getcwd())

from pret.main import extract_js_dependencies

app = typer.Typer()


@app.command()
def prepack(stub_module: str, output_path: str, cwd: bool = True):
    """
    Pre-packs a stub module into a single javascript file.

    Parameters
    ----------
    stub_module
        The name of the stub module to prepack.
    output_path
        The path to the output file.
    cwd
        Whether to add the current working directory to the python path.
    """

    output_path = Path(output_path)

    if cwd:
        sys.path.insert(0, os.getcwd())

    module = __import__(stub_module, fromlist=["*"])
    print(f"Pre-pack from {stub_module} at {module.__file__}")

    pickler = PretMarshaler(allow_error=True)
    pickler.visit(module)
    js_globals_file_str = extract_js_dependencies(pickler.accessed_global_refs)[0]

    output_path.parent.mkdir(parents=True, exist_ok=True)
    with output_path.open("w") as f:
        f.write(js_globals_file_str)


if __name__ == "__main__":
    app()



import os
import subprocess

from org.transcrypt import utils

closureCompilerPath = '{}/closure_compiler/compiler.jar'.format (os.path.dirname (os.path.abspath (__file__)) .replace ('\\', '/'))

# Minifier has to accept JavaScript 6 input code, it is there in the autotest, even if not executed.
def run (targetDir, sourceFileName, targetFileName, mapFileName = None, prettify = False):
    params = [
        'java', '-jar',
        closureCompilerPath,
        '--language_out=ECMASCRIPT6_STRICT',
        '--compilation_level', 'WHITESPACE_ONLY',
        '--js', sourceFileName,
        '--js_output_file', targetFileName
    ]
    
    if prettify:
        params += ['--formatting', 'PRETTY_PRINT']
    
    if utils.commandArgs.map and not prettify:
        params += [
            '--create_source_map', mapFileName,
            '--source_map_format=V3'
        ]

    origDir = os.getcwd ()
    os.chdir (targetDir)    # So the map will store sourcePath and targetPath as filenames rather than full paths
    subprocess.run (params)
    os.chdir (origDir)
    


import os
import sys
import argparse
import inspect
import tokenize
import re
import time

class Any:
    def __init__ (self, **attribs):
        for attrib in attribs:
            setattr (self, attrib, attribs [attrib])

defaultJavaScriptVersion = 5

class CommandArgsError (BaseException):
    pass
    
class CommandArgsExit (BaseException):
    pass
    
class ArgumentParser (argparse.ArgumentParser):
    def error (self, message):
        self.print_help (sys.stdout)
        if message:
            log (True, '\nError: {}\n', message)
        raise CommandArgsError ()
        
    def exit (self, status = 0, message = None):
        if message:
            log (True, 'Exit: {}', message)
        raise CommandArgsExit ()

class CommandArgs:
    def parse (self):
        self.argParser = ArgumentParser ()
        
        self.argParser.add_argument ('source', nargs='?', help = ".py file containing source code of main module")
        self.argParser.add_argument ('-a', '--anno', help = "annotate target files that were compiled from Python with source file names and source line numbers", action = 'store_true')
        self.argParser.add_argument ('-am', '--alimod', help = "use aliasing for module paths", action = 'store_true')
        self.argParser.add_argument ('-b', '--build', help = "rebuild all target files from scratch", action = 'store_true')
        self.argParser.add_argument ('-c', '--complex', help = "enable complex number support, locally requires operator overloading", action = 'store_true')
        self.argParser.add_argument ('-d', '--docat', help = "enable __doc__ attributes. Apply sparsely, since it will make docstrings part of the generated code", action = 'store_true')
        self.argParser.add_argument ('-da', '--dassert', help = "debug: activate assertions", action = 'store_true')
        self.argParser.add_argument ('-dc', '--dcheck', help = "debug: perform lightweight consistency check", action = 'store_true')
        self.argParser.add_argument ('-de', '--dextex', help = "debug: show extended exception reports", action = 'store_true')
        self.argParser.add_argument ('-dl', '--dlog', help = "debug: log compiler messages to disk", action = 'store_true')
        self.argParser.add_argument ('-dm', '--dmap', help = "debug: dump human readable source map", action = 'store_true')
        self.argParser.add_argument ('-dn', '--dnostrip', help = "debug: no comment stripping of __core__ and __builtin__ in-line modules", action = 'store_true')
        self.argParser.add_argument ('-ds', '--dstat', help = "debug: validate static typing using annotations", action = 'store_true')
        self.argParser.add_argument ('-dt', '--dtree', help = "debug: dump syntax tree", action = 'store_true')
        self.argParser.add_argument ('-e', '--esv', nargs='?', help = "ecma script version of generated code, default = 6. The symbol __esv<versionnr>__ is added to the global symbol list, e.g. __esv7__.")
        self.argParser.add_argument ('-ec', '--ecom', help = "enable executable comments, seen as comments by CPython but as executable statements by Transcrypt", action = 'store_true')
        self.argParser.add_argument ('-f', '--fcall', help = "enable fastcall mechanism by default. You can also use __pragma__ ('fcal') and __pragma__ (\'nofcall\')", action = 'store_true')
        self.argParser.add_argument ('-g', '--gen', help = "enable generators and iterators. Disadvised, since it will result in a function call for each loop iteration. Preferably use __pragma__ ('gen') and __pragma__ ('nogen')", action = 'store_true')
        self.argParser.add_argument ('-i', '--iconv', help = "enable automatic conversion to iterable by default. Disadvised, since it will result in a type check for each for-loop. Preferably use __pragma__ ('iconv') and __pragma__ (\'noiconv\') to enable automatic conversion locally", action = 'store_true')
        self.argParser.add_argument ('-jc', '--jscall', help = "enable native JavaScript calls for Python methods. This is fast, but doesn't support bound method assignment, decorators and non-instance methods. Preferably use __pragma__ ('jscall') and __pragma__ ('nojscall') to enable native JavaScript calls locally", action = 'store_true')
        self.argParser.add_argument ('-jk', '--jskeys', help = "interpret {key: 'value'} as {'key': 'value'} and forbid {key (): 'value'}, as JavaScript does. Disadvised, since it's less flexible than the Python interpretation. Either follow Python semantics by using {'key': 'value'} explicitly if you want literal keys or use __pragma__ ('jskeys') and __pragma__ ('nojskeys') locally instead to make clear local deviation from Python semantics", action = 'store_true')
        self.argParser.add_argument ('-jm', '--jsmod', help = "give %% and %%= JavaScript rather than Python behaviour. Disadvised, since it deviates from the mathematical 'modulo' operator. Either follow Python semantics or use __pragma__ ('jsmod') and __pragma__ ('nojsmod') locally instead to make clear local deviation.", action = 'store_true')
        self.argParser.add_argument ('-k', '--kwargs', help = "enable keyword arguments by default. In general this is disadvised, use __pragma__ ('kwargs') and __pragma__('nokwargs') locally instead to prevent bloated code", action = 'store_true')
        self.argParser.add_argument ('-kc', '--keycheck', help = "enable checking for existence of dictionary keys. In general this is disadvised, use __pragma__ ('keycheck') and __pragma__('nokeycheck') locally instead to prevent bloated code", action = 'store_true')
        self.argParser.add_argument ('-l', '--license', help = "show license", action = 'store_true')
        self.argParser.add_argument ('-m', '--map', help = "generate source map", action = 'store_true')
        self.argParser.add_argument ('-n', '--nomin', help = "no minification", action = 'store_true')
        self.argParser.add_argument ('-o', '--opov', help = "enable operator overloading by default. In general this is disadvised, use __pragma__ ('opov') and __pragma__('noopov') locally instead to prevent slow code", action = 'store_true')
        self.argParser.add_argument ('-p', '--parent', nargs='?', help = "object that will hold application, default is window. Use -p .none to generate orphan application, e.g. for use in node.js")
        self.argParser.add_argument ('-r', '--run', help = "run source file rather than compiling it", action = 'store_true')
        self.argParser.add_argument ('-s', '--symbols', nargs='?', help = "names, joined by $, separately passed to main module in __symbols__ variable")
        self.argParser.add_argument ('-sf', '--sform', help = "enable support for string formatting mini language", action = 'store_true')
        self.argParser.add_argument ('-t', '--tconv', help = "enable automatic conversion to truth value by default. Disadvised, since it will result in a conversion for each boolean. Preferably use __pragma__ ('tconv') and __pragma__ (\'notconv\') to enable automatic conversion locally", action = 'store_true')
        self.argParser.add_argument ('-u', '--unit', nargs='?', help = "compile to units rather than to monolithic application. Use -u .auto to autogenerate dynamically loadable native JavaScript modules, one per Python module. Use -u .run to generate the loader and the staticcally loadable runtime unit. Use -u .com to generate a statically loadable component unit.")
        self.argParser.add_argument ('-v', '--verbose', help = "show all messages", action = 'store_true')
        self.argParser.add_argument ('-x', '--x', help = "reserved for extended options")
        self.argParser.add_argument ('-xr', '--xreex', help = "re-export all imported names", action = 'store_true')
        self.argParser.add_argument ('-xg', '--xglobs', help = "allow use of the 'globals' function", action = 'store_true')
        self.argParser.add_argument ('-xp', '--xpath', nargs = '?', help = "additional module search paths, joined by $, #'s will be replaced by spaces")
        self.argParser.add_argument ('-xt', '--xtiny', help = "generate tiny version of runtime, a.o. lacking support for implicit and explicit operator overloading. Use only if generated code can be validated, since it will introduce semantic alterations in edge cases", action = 'store_true')
        self.argParser.add_argument ('-*', '--star', help = "Like it? Grow it! Go to GitHub and then click [* Star]", action = 'store_true')
        
        self.projectOptions = self.argParser.parse_args () .__dict__
        self.__dict__.update (self.projectOptions)
        
        # Signal invalid switches
        
        def logAndExit (message):
            log (True, message)
            sys.exit (1)
        
        invalidCombi = 'Invalid combination of options'
        
        if not (self.license or self.star or self.source):
            logAndExit (self.argParser.format_usage () .capitalize ())
        elif self.map and self.unit:
            logAndExit ('{}: -m / --map and -u / --unit'.format (invalidCombi))   
        elif self.parent and self.unit == '.com':
            logAndExit ('{}: -p / --parent and -u / --unit .com'.format (invalidCombi))
        elif self.parent == '.export' and self.esv and int (self.esv) < 6:
            logAndExit ('{}: -p / --parent .export and -e / --esv < 6'.format (invalidCombi))        
        elif self.unit == '.auto' and self.esv and int (self.esv) < 6:
            logAndExit ('{}: -u / --unit .auto and -e / --esv < 6'.format (invalidCombi))
        
        # Set dependent switches
        
        # (for future use)
            
        # Correcting line counts for source map
            
        global extraLines
        extraLines = [
            # Make identifier __pragma__ known to static checker
            # It was only known in JavaScript from __core__.js, which the checker doesn't see
            # __ pragma__ ('<all>') in JavaScript requires it to remain a function, as it was in the core
            # It can't be skipped, since it has to precede __pragma__ ('skip'), to make the checker accept that
            'def __pragma__ (): pass',
        
            # Make __include__ known to the static checker
            '__pragma__ (\'skip\')',            
            '__new__ = __include__ = 0',    
            '__pragma__ (\'noskip\')',
            ''
        ] if commandArgs.dcheck else []
        global nrOfExtraLines
        nrOfExtraLines = max (len (extraLines) - 1, 0)  # Last line only serves to force linefeed
        extraLines = '\n'.join (extraLines)
        
commandArgs = CommandArgs ()
   
def create (path, binary = False):
    for i in range (10):
        try:
            os.makedirs (os.path.dirname (path), exist_ok = True)
            
            if binary:
                return open (path, 'wb')
            else:
                return open (path, 'w', encoding = 'utf-8')
                
            if i > 0:
                log (True, f'Created {path} at attempt {i + 1}')
                
        except:
            time.sleep (0.5)
    else:
        raise Error (f'Failed to create {path}')
        

def tryRemove (filePath):
    try:
        os.remove (filePath)
    except:
        pass
    
def formatted (*args):  # args [0] is string, args [1 : ] are format params
    try:
        return str (args [0]) .format (*args [1 : ])
    except IndexError:  # Tuple index out of range in format tuple
        return ' '.join (args)
   
logFileName = 'transcrypt.log'  # ... Use envir.transpiler_name
   
try:
    os.remove (logFileName)
except: # Assume logfile doesn't yet exist
    pass
   
def log (always, *args):
    if always or commandArgs.verbose:
        print (formatted (*args), end = '')
        try:
            if commandArgs.dlog:
                with open (logFileName, 'a') as logFile:
                    logFile.write (formatted (*args))
         
        except: # Assume too early, commandArgs not yet set
            pass
            
program = None
def setProgram (aProgram):
    global program
    program = aProgram
         
class Error (Exception):
    # First error encountered counts, for all fields, because it's closest to the cause of trouble
    # One error at a time, just like Python, clear and simple
    # The error object contains the import trail, the error line number and the error message

    def __init__ (self, lineNr = 0, message = ''):
        self.lineNr = lineNr - nrOfExtraLines
        self.message = message
        # The name of the module and of its import trail is known from the import stack, so no need to pass it as a parameter
        
    def set (self, lineNr = 0, message = ''):          
        if not self.lineNr:
            self.lineNr = lineNr - nrOfExtraLines
            
        if not self.message:
            self.message = message
            
    def __str__ (self):
        result = 'Error while compiling (offending file last):'

        # Successively report each module in the import trail, "oldest" first, except the one that caused the error
        for importRecord in program.importStack [ : -1]:
            try:
                sourcePath = importRecord [0] .sourcePath
            except:
                sourcePath = '<unknown>'
            result += '\n\tFile \'{}\', line {}, at import of:'.format (sourcePath, importRecord [1])
        
        # After that, report the module and line that caused the error
#        result += '\n\tFile \'{}\', line {}, namely:'.format (str (program.importStack [-1][0] .sourcePath), self.lineNr)
        result += '\n\tFile \'{}\', line {}, namely:'.format (str (program.importStack [-1][0] .name), self.lineNr)
        
        # And, lastly, report the error message
        result += '\n\t{}'.format (self.message)
        return result
        
def enhanceException (exception, **kwargs):
    # If not all required info, such as a line number, is available at the location where the exception was raised,
    # it is enhanced later on by adding this missing info at the earliest occasion

    if isinstance (exception, Error):
        exception.set (**kwargs)
        result = exception
    else:
        result = Error (**kwargs)
    
    if commandArgs.dextex:
        print ('''
    Exception of class {0} enhanced at:
        file: {1}
        function: {3}
        line: {2}
        context: {4}
        kwargs: {5}
        result: {6}
    '''.format (exception.__class__, *inspect.stack () [1][1:-1], kwargs, result))

    raise result from None
        
def digestJavascript (code, symbols, mayStripComments, mayRemoveAnnotations, refuseIfAppearsMinified = False):
    '''
    - Honor ifdefs
    - Strip comments if allowed by commend line switch AND indicated by pragma
    - Harvest import and export info
    '''

    if refuseIfAppearsMinified and code [0] != '/':
        return None
    
    stripComments = False
        
    def stripSingleLineComments (line):
        pos = line.find ('//')
        return (line if pos < 0 else line [ : pos]) .rstrip ()

    passStack = []

    def passable (targetLine):        
        # Has to count, since comments may be inside ifdefs
        
        nonlocal stripComments
        
        def __pragma__ (name, *args):

            nonlocal stripComments
        
            if name == 'stripcomments':
                stripComments = mayStripComments
            if name == 'ifdef':
                passStack.append (args [0] in symbols)
            elif name == 'ifndef':
                passStack.append (not args [0] in symbols)
            elif name == 'else':
                passStack [-1] = not passStack [-1]
            elif name == 'endif':
                passStack.pop ()

        strippedLine = targetLine.lstrip ()
        if stripComments and strippedLine.startswith ('/*'):
            passStack.append (False)
            return all (passStack)  # So skip this line
        elif stripComments and strippedLine.endswith ('*/'):
            passStack.pop ()        # Possibly pass next line
        elif strippedLine.startswith ('__pragma__') and (
            'stripcomments' in strippedLine or
            'ifdef' in strippedLine or
            'ifndef' in strippedLine or
            'else' in strippedLine or
            'endif' in strippedLine
        ):
            exec (strippedLine)
            return False            # Skip line anyhow, independent of passStack
        else:
            return all (passStack)  # Skip line only if not in passing state according to passStack
    
    passableLines = [line for line in code.split ('\n') if passable (line)]
    
    if stripComments:
        passableLines = [commentlessLine for commentlessLine in [stripSingleLineComments (line) for line in passableLines] if commentlessLine]
        
    result = Any (
        digestedCode = '\n'.join (passableLines),
        nrOfLines = len (passableLines),
        exportedNames = [],
        importedModuleNames = []
    )
    
    namesPattern = re.compile ('({.*})')
    pathPattern = re.compile ('([\'|\"].*[\'|\"])')
    wordPattern = re.compile (r'[\w$]+')
    for line in passableLines:
        words = wordPattern.findall (line)
        
        if words:
            if mayRemoveAnnotations and words [0] == '/*':  # If line starts with an annotation
                words = words [3 : ]                        # Remove the annotation before looking for export / import keywords
                
            if words [0] == 'export':
                # Deducing exported names from JavaScript is needed to facilitate * import by other modules
                
                if words [1] in {'var', 'function'}:
                    # Export prefix:    "export var ... or export function ..."
                    
                    result.exportedNames.append (words [2])
                else:
                    # Transit export:   "export {p, q, r, s};"  
                    
                    # Find exported names as "{p, q, r, s}"
                    match = namesPattern.search (line)
                    
                    # Substitute to become "{'p', 'q', 'r', 's'}" and use that set to extend the exported names list
                    if match:
                        result.exportedNames.extend (eval (wordPattern.sub (lambda nameMatch: f'\'{nameMatch.group ()}\'', match.group (1))))
                     
            if words [0] == 'import':
                # Deducing imported modules from JavaScript is needed to provide the right modules to JavaScript-only modules
                # They may have an explicit import list for unqualified access or an import * for qualified access
                # In both cases only the path of the imported module is needed, to be able to provide that module
                # It can be a path without extension, allowing both .py and .js files as imported modules
                #
                # - Unqualified import:   "import {p, q as Q, r, s as S} from '<relative module path>'"
                # - Qualified import:     "import * from '<relative module path>'"  
                
                match = pathPattern.search (line)
                if match:
                    result.importedModuleNames.append (eval (match.group (1)) [2:-3])
                
    return result

    


import os
import subprocess
from pathlib import Path

import typer

app = typer.Typer()


def get_current_pyproject_name():
    path = Path.cwd()
    while not (path / "pyproject.toml").exists():
        if path == path.parent:
            raise FileNotFoundError("pyproject.toml not found")
        path = path.parent
    # I don't want to add another dependency just for this
    text = (path / "pyproject.toml").read_text()
    project_section = text[text.index("[project]") : text.index("\n[")]
    name_idx = project_section.index("name")
    name = project_section[name_idx:].split("\n")[0].split("=")[1].strip().strip('"')
    return name


@app.command()
def stub(path_to_js_module: str, global_name: str, output_path: str):
    script_path = Path(__file__).parent / "generate-py-stubs.js"
    env = os.environ.copy()
    env["NODE_PATH"] = str(Path.cwd())
    pyproject_name = get_current_pyproject_name()
    # subprocess.run(["yarn", "add", "typescript@4.5.5", "--dev"], check=True)
    print(path_to_js_module, pyproject_name, global_name, output_path)
    subprocess.run(
        [
            "node",
            script_path,
            path_to_js_module,
            pyproject_name,
            global_name,
            output_path,
        ],
        env=env,
        check=True,
    )


if __name__ == "__main__":
    app()



import pkgutil

__path__ = pkgutil.extend_path(__path__, __name__)

from . import ipython_var_cleaner  # noqa: F401
from .main import run
from .render import component
from .store import create_store
from .hooks import (
    use_callback,
    use_effect,
    use_memo,
    use_ref,
    use_state,
    use_store_snapshot,
    use_body_style,
    use_event_callback,
)
from .manager import server_only

__version__ = "0.4.1"

__all__ = [
    "component",
    "create_store",
    "run",
    "server_only",
    "use_callback",
    "use_effect",
    "use_memo",
    "use_ref",
    "use_state",
    "use_store_snapshot",
    "use_body_style",
    "use_event_callback",
]



# First run a test from the command prompt, generating an HTML file.
# The output of the test is stored in a DIV.
# Also the script is automatically included in the HTML file.
# Loading the HTML file will run the script.
# This will compare the output of the script running in the browswer to the output in the DIV.
# If those two match, the test reports OK, else it reports failure.


from org.transcrypt.stubs.browser import __main__, __envir__, __pragma__
from org.transcrypt.autotester.html import HTMLGenerator, DataConverter, JSTesterUI, itemsAreEqual

# Don't import __envir__ from __base__ since it will overwrite __buildin__.__envir__ in the browser
# Import from stubs will be skipped in the browser
# ... The ice is a bit thin here

__pragma__ ('nokwargs')

import itertools

def getFileLocation(ancestor):
    """ This function needs to crawl up the stack
    and find out where the ancestor caller of
    this function was in the source code of either the
    python or javascript, depending on environment.
    @param ancestor the ancestor of this function that
      we want to capture file information about.
    @return string indicating the file position and line number
    """
    if __envir__.executor_name == __envir__.transpiler_name: # js
        s = None
        __pragma__('js', '{}',
            '''
            var e = new Error();
            if ( ! e.stack ) {
              console.log("MAJOR ISSUE: Browser Error lacks Stack");
            } else {
              s = e.stack;
            }
            ''')
        # Now we will process the stack to find the grandparent
        # calling function
        # @note - I'm explicitly not including a 're' module
        #    dependency here
        frames = None
        __pragma__('js', '{}',
            '''
            var linereg = new RegExp("\\n\\r|\\n", "g");
            frames = s.toString().split(linereg);
            ''')
        if ( frames is None or (len(frames) < 2)):
            __pragma__('js', '{}', 'console.log("Failed to Split Stack");')
            return("UNKNOWN:???")

        # @note - if the call stack in transcrypts javascript
        #   translation changes then this index may need to change
        #   @todo - need more work here to determine this because
        #     this is fragile
        gpFrame = frames[(ancestor*2 + 1)]
        # This regex splits the string coming from the javascript
        # stacktrace so that we can connect the file and line number
        # runTests (http://localhost:8080/run/autotest.js:3159:8)
        #  func       URL             filename    lineno:colno
        # Group 1 = function
        # Group 2 & 3 = protocol and hostname
        # Group 4 = Path on this host (filename is at the end)
        # Group 5 = lineno
        # Group 6 = column number in file
        frameReg = r"([^(]*)\(?([^:]*:)\/{2,3}([^:/]*:?)([^:]*):(\d+):(\d+)"
        m = None
        __pragma__('js', '{}',
            '''
            var r = new RegExp(frameReg);
            m = r.exec(gpFrame);
            ''')
        if m:
            filepath = m[4]
            # Split the filepath and take the last element
            # to the get filename
            pathParts = filepath.split("/")
            filename = pathParts[len(pathParts)-1]
            lineno = m[5]
            return( "{}:{}".format(filename, lineno) )
        else:
            __pragma__('js', '{}', 'console.log("Failed to Match Frame", gpFrame);')
            return("UNKNOWN:???")
    #ELSE
    # Needed because Transcrypt imports are compile time
    __pragma__("skip")
    from inspect import getframeinfo, stack
    s = stack()
    caller = getframeinfo(s[ancestor][0])
    # Trim the file name path so that we don't get
    # a lot of unnecessary content
    filepath = caller.filename
    # @todo - this is a hack - we should use os.path
    pathParts = filepath.split('/')
    filename = "/".join(pathParts[-2:])
    return( "%s:%d" % (filename, caller.lineno))
    __pragma__ ('noskip')



class AutoTester:
    """ Main testing class for comparing CPython to Transcrypt. This
    class is primarily used by calling the "check" method to confirm that
    the result is the same in both environments and "done" when all checks
    for a particular module have been completed.
    """
    def __init__ (self, symbols = []):
        self.symbols = symbols
        # refDict/testDict contains the test results
        # of each testlet identified by name as the key
        self._currTestlet = "UNKNOWN"
        self.testDict = {}
        self.refDict = {}

        if __envir__.executor_name == __envir__.transpiler_name:
            self.ui = JSTesterUI()
        else:
            self.ui = None

    def sortedRepr (self, any):
        # When using sets or dicts, use elemens or keys
        # of one type, in sort order
        def tryGetNumKey (key):
            if type (key) == str:   # Try to interpret key as numerical, see comment with repr function in __builtins__
                try:
                    return int (key)
                except:
                    try:
                        return float (key)
                    except:
                        return key
            else:
                return key

        if type (any) == dict:
            return '{' + ', '.join ([
                '{}: {}'.format (repr (key), repr (any [key]))
                for index, key in enumerate (sorted ([tryGetNumKey (key) for key in any.keys ()], key = lambda aKey: str (aKey)))
            ]) + '}'
        elif type (any) == set:
            if len (any):
                return '{' + ', '.join (sorted ([str (item) for item in list (any)])) + '}'
            else:
                return repr (any)
        elif type (any) == range:
            return repr (list (any))
        else:
            return repr (any)

    __pragma__('kwargs')
    def check (self, *args, ancestor = 2):
        """ Given a set of values from either the python or transcrypt
        environments, we log the position of the check call in the test
        and representative values of the passed arguments for later
        comparison.
        """
        position=getFileLocation(ancestor)
        # N.B. stubs.browser provides a special sorting repr
        item = ' '.join ([self.sortedRepr (arg) for arg in args])
        if __envir__.executor_name == __envir__.transpiler_name:
            self.testDict[self._currTestlet].append((position,item))
        else:
            self.refDict[self._currTestlet].append((position,item))

    __pragma__('nokwargs')

    def expectException(self, func):
        """ This method attempts to call the passed method and
        checks to see whether an exception was generated.
        @return string indicating "no exception" or "exception"
        """
        try:
            func()
            return("no exception")
        except Exception as exc:
            return("exception")

    def throwToError(self, func):
        """ This function invokes the passed function and then
        converts an exception to an error response so that
        the unit test can continue even in the case where an
        exception may or may not occur.
        """
        try:
            return(func())
        except Exception as exc:
            return (None, "!!!{}".format(str(exc)))

    def checkEval(self, func):
        """ Check the result of the passed function which is
        invoked without arguments. If this function throws an
        exception, that exception is caught and converted to an error
        with can be compared against the result. This allows the
        user to control for exception that may or may not be generated
        in the unit tests
        """
        ret = self.throwToError(func)
        self.check(ret, ancestor = 3)


    def checkPad(self, val, count):
        """ This method is to help manage flow control in unit tests and
        keep all unit tests aligned
        """
        for i in range(0, count):
            self.check(val)

    def _getTotalErrorCnt(self, testData, refData):
        """ This method determines the total number of non-matching
        values in the test and reference data for a particular module.
        """
        errCount = 0
        for i,(refPos, refItem) in enumerate(refData):
            try:
                testPos,testItem = testData[i]
                if not itemsAreEqual (testItem, refItem):
                    errCount+=1
            except:
                errCount+=1
        return(errCount)

    def compare (self):
        # Load the python reference data from the hidden HTML div
        dc = DataConverter()
        self.refDict = dc.getPythonResults()

        totalErrors = 0
        sKeys = sorted(self.refDict.keys())
        for key in sKeys:
            refData = self.refDict[key]
            try:
                testData = self.testDict[key]
                if ( testData is None ):
                    raise KeyError("No Test Data Module: {}".format(key))
            except KeyError:
                # No Test Data found for this key - we will populate with
                # errors for all ref data
                self.ui.appendSeqRowName(key, len(refData))
                for i,(refPos, refItem) in enumerate(refData):
                    self.ui.appendTableResult(key, None, None, refPos, refItem, False)
                continue
            # know we have testData so let's determine the total number of
            # errors for this test module. This will allow us to both set
            # the num of errors in the test module header row and set the
            # rows to the appropriate initial collapsed/expanded state.
            errCount= self._getTotalErrorCnt(testData, refData)
            collapse = (errCount == 0)
            self.ui.appendSeqRowName(key, errCount)

            # Now we will populate the table with all the rows
            # of data fro the comparison
            for i,(refPos, refItem) in enumerate(refData):
                try:
                    # This will throw if testData's length is
                    # shorter than refData's
                    testPos,testItem = testData[i]
                except:
                    testPos = None
                    testItem = None

                self.ui.appendTableResult(
                    key, testPos, testItem, refPos, refItem, collapse
                )

            totalErrors += errCount

        self.ui.setOutputStatus( totalErrors == 0 )


    def _cleanName(self, name):
        """ Clean the passed name of characters that won't be allowed
            in CSS class or HTML id strings.
        """
        # Convert testletName to replace any of the characters that
        # are not acceptable in a CSS class or HTML id - this is to
        # make our lives easier
        # @note - I'm SPECIFICALLY not using a regex here because the
        #   regex engine module is still under dev and could possibly
        #   have issues
        ret = name
        invalidChars = [
            '~', '!', '@', '$', '%',
            '^', '&', '*', '(', ')',
            '+', '=', ',', '.', '/',
            "'", ';', ':', '"', '?',
            '>', '<', '[', ']', '\\',
            '{', '}', '|', '`', '#',
            " ",
        ]
        for ch in invalidChars:
            ret = ret.replace(ch, "_")
        return(ret)

    def run (self, testlet, testletName):
        testletName = self._cleanName(testletName)
        self._currTestlet = testletName
        if __envir__.executor_name == __envir__.transpiler_name:
            self.testDict[self._currTestlet] = []
        else:
            self.refDict[self._currTestlet] = []
        try:
            testlet.run (self)
        except Exception as exc:
            if ( self.ui is not None ):
                self.ui.setOutputStatus(False)
                self.ui.showException(testletName, exc)
            else:
                # Error - No UI yet, reraise specific exception to enable finding out why
                raise


    def done (self):
        if __envir__.executor_name == __envir__.transpiler_name:
            self.compare ()
        else:
            fnameBase = __main__.__file__.replace ('\\', '/')
            hg = HTMLGenerator(fnameBase)
            hg.generate_html(self.refDict)
